
/* Copyright (C) 2001 Willem Robert van Hage
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, you can get it from http://www.fsf.org
 * or write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
 */

#include "map.h"

Map *map = NULL; /* the current map */

/*
 * MAP FUNCTIONS
 * (for space functions see below)
 */

Map *map_new(void) {
  Map *map = (Map*)malloc(sizeof(Map)*1);
  (map)->spaces = g_hash_table_new(g_direct_hash,g_direct_equal);
  return map;
}

void map_destroy(Map **map) {
  if((map == NULL) || (*map == NULL)) return; /* remember lazy evaluation? */
  if((*map)->spaces != NULL) (void)map_remove_all_spaces(*map);
  g_free(*map);
  *map = NULL;  
}

/* a bunch of accessor functions for the GLIB hashtable map->spaces */
void map_add_space(Map *map,Space *space) {
  g_hash_table_insert(map->spaces,space_to_key(space),space);
}
void map_remove_space(Map *map,Space *space) {
  g_hash_table_remove(map->spaces,space_to_key(space));
}
void map_remove_space_xy(Map *map,int x,int y) {
  map_remove_space(map,xy_to_key(x,y));
}
guint map_remove_all_spaces(Map *map) {
  return g_hash_table_foreach_remove(map->spaces,space_remove_func,NULL);
}
Space *map_get_space(Map *map,int x,int y) {
  return (Space*)g_hash_table_lookup(map->spaces,xy_to_key(x,y));
}
gpointer xy_to_key(int x, int y) {
  /* take 16 bits of both ints and glue them together to form a 32 bit int */
  return GINT_TO_POINTER((x & 0xFFFF) | ((y & 0xFFFF) << 16));
}
gpointer space_to_key(Space *space) {
  /* two references (->) are quicker than a function call to xy_to_key() */
  return GINT_TO_POINTER((space->x & 0xFFFF) | ((space->y & 0xFFFF) << 16));
}
/* when destroying the hash table destroy the spaces it holds too... */
gboolean space_remove_func(gpointer key,gpointer value, gpointer data) {
  space_destroy((Space**)&value);
  return TRUE; /* yes, go ahead and remove it, *grumble* */
}


/* a temporary function to generate a test map */
void map_generate(Map *map) {

  int i,j;

  srand(time(NULL));

  for(j=0;j<=screen_h/tile_size;j++) /* for all spaces you can see */
  for(i=0;i<=screen_w/tile_size;i++) {
    Space *space = space_new(i,j);
    ((int*)space->data[SPACE_DATA_BOTTOM_IMAGE])[0] = load_image("grass1.png");
    space_frame_set(space,SPACE_DATA_BOTTOM_IMAGE,
		    rand()%2,rand()%2,2,2,0,rand()%5);
    map_add_space(map,space);
    add_update_coord(i,j); /* remember to draw this space on screen */
  }
  { /* create one knight */
    Mob *knight = (Mob*)mob_new("Sir Strandarr",9,9);
    for(j=0;j<2;j++)
    for(i=0;i<2;i++)  
      space_mob_add(knight,map_get_space(map,9+i,9+j));
    ((int*)knight->data[MOB_DATA_IMAGE])[0] = load_image("knight1.png");
    mob_frame_set(knight,0,0,2,2,2,0);
    player = knight;
    knight->data[MOB_DATA_BEHAVIOR] = (void*)load_behavior("player_behavior");
    /* no need to call add_update_coord(), because we already did for (5,5) */
  }
  { /* create one ogre */
    Mob *ogre = (Mob*)mob_new("Krunk the ogre",18,18);
    for(j=0;j<4;j++)
    for(i=0;i<4;i++)  
      space_mob_add(ogre,map_get_space(map,18+i,18+j));
    ((int*)ogre->data[MOB_DATA_IMAGE])[0] = load_image("ogre1.png");
    mob_frame_set(ogre,0,0,4,4,2,0);
    ogre->data[MOB_DATA_BEHAVIOR] = (void*)load_behavior("walk_test2");
  }
  for(j=0;j<6;j++) /* for the-size-of-the-tree (3,3) */
  for(i=0;i<6;i++) {
    Space *space = map_get_space(map,i+2,j+2);
    /* the trunk, on the bottom right is in the middle layer,
     * the rest of the tree, leaves etc. in the top layer...
     */
    if((i>=4)&&(j>=4)) {
      ((int*)space->data[SPACE_DATA_MIDDLE_IMAGE])[0] =
	load_image("tree1.png");
      space_frame_set(space,SPACE_DATA_MIDDLE_IMAGE,i,j,6,6,0,0);
      SET_FLAG(*space->flags,SPACE_FLAGS_SOLID);
    } else {
      ((int*)space->data[SPACE_DATA_TOP_IMAGE])[0] = 
	load_image("tree1.png");
      space_frame_set(space,SPACE_DATA_TOP_IMAGE,i,j,6,6,0,0);
    }
  }
  set_viewport(0,0,(screen_w/tile_size)+1,(screen_h/tile_size)+1);  

}



/*
 * SPACE FUNCTIONS
 */

/* a Space is roughly 25+ sizeof(int)'s, so that's 100+ bytes or so. */
Space *space_new(int x, int y) {

  Space *space = (Space*)malloc(sizeof(Space)*1);
  space->x = x;
  space->y = y;
  
  /* SpaceFlags is an enum. of bit offsets, devide it by 8 and you got bytes */
  space->flags = (Uint8*)malloc((SPACE_FLAGS_N/8)+1);
  memset(space->flags,0,(SPACE_FLAGS_N/8)+1);

  space->data = (void**)malloc(sizeof(void*)*SPACE_DATA_N);
  memset(space->data,0,sizeof(void*)*SPACE_DATA_N);
  
  /* the seven numbers we need to indicate a tile... */
  (int)space->data[SPACE_DATA_BOTTOM_IMAGE] = (int*)malloc(sizeof(int)*7);
  (int)space->data[SPACE_DATA_MIDDLE_IMAGE] = (int*)malloc(sizeof(int)*7);
  (int)space->data[SPACE_DATA_TOP_IMAGE]    = (int*)malloc(sizeof(int)*7);
  { /* memset(,!0,) doesn't work right for elements of sizeof() > 1 */
    int i;
    for(i=0;i<7;i++) ((int*)space->data[SPACE_DATA_BOTTOM_IMAGE])[i] = -1;
    for(i=0;i<7;i++) ((int*)space->data[SPACE_DATA_MIDDLE_IMAGE])[i] = -1;
    for(i=0;i<7;i++) ((int*)space->data[SPACE_DATA_TOP_IMAGE])[i]    = -1;
  }
  /* the list of movable objects is a NULL terminated array of pointers */
  space->data[SPACE_DATA_MOBS] = (Mob**)malloc(sizeof(Mob*)*1);
  ((Mob**)space->data[SPACE_DATA_MOBS])[0] = NULL;

  return space;

}

void space_destroy(Space **space) {

  if((space == NULL) || (*space == NULL)) return;
  if((*space)->data != NULL) {
    if((*space)->data[SPACE_DATA_BOTTOM_IMAGE] != NULL)
      free((*space)->data[SPACE_DATA_BOTTOM_IMAGE]);
    if((*space)->data[SPACE_DATA_MIDDLE_IMAGE] != NULL)
      free((*space)->data[SPACE_DATA_MIDDLE_IMAGE]);
    if((*space)->data[SPACE_DATA_TOP_IMAGE] != NULL)
      free((*space)->data[SPACE_DATA_TOP_IMAGE]);

    if((*space)->data[SPACE_DATA_MOBS] != NULL)
      free((*space)->data[SPACE_DATA_MOBS]);

    free((*space)->data);
  }
  if((*space)->flags != NULL) free((*space)->flags);
  free(*space);
  *space = NULL;

}

